#define PRGNAME "AGInfo"
#define PRGVERSION "1.13"
#define PRGDATE "20020331"

/* #define DEBUG */

/* define LARGE for Amiga, DJGPP etc. */
/* #define LARGE */

/*
 *  aginfo.c
 *
 *  - detects AGI v2/v3 version & game version
 *
 *  Author: Paul David Doherty <agi@pdd.de>
 *
 *  v0.9:    7 Jul 1999
 *  v1.0:    18 Jul 1999  not much changed
 *  v1.01:   17 Aug 1999  MH1/Amiga suppt; hidden list/wanted options
 *  v1.02:   24 Nov 1999  more list options
 *  v1.03:   25 Nov 1999  more possible xxDIR names (MGDIR, KQ3DIR, PQ1DIR)
 *  v1.10:    8 May 2001  new version number for millennium shift :-)
 *  v1.11:   26 Nov 2001  small typecast fix (thx to Miron Schmidt)
 *  v1.12:   28 Mar 2002  fixed bug in bestmatch algorithm
 *  v1.13:   31 Mar 2002  prepared for source code release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#ifdef __TURBOC__
#include <io.h>
#endif

/* typedefs needed by agidata.h */

typedef unsigned char type8;
typedef unsigned short type16;
typedef unsigned long int type32;

#ifdef LARGE
typedef type32 type16_32;
#define CUTOFF 100000L
#else
typedef type16 type16_32;
#define CUTOFF 64000L
#endif

#include "agidata.h"
#include "agicrc.h"

/* other global stuff */
#define TRUE 1
#define FALSE 0

FILE *fdi;
type8 *infiledata;
type32 infilelength;

type16 crc_checksum = 0;
type32 df_checksum;

/* Amiga version string */
char *amiverstag = "\0$VER: " PRGNAME " " PRGVERSION " (" PRGDATE ")";

/*>>>>>>>>>>>>>>>>>>>>> various routines <<<<<<<<<<<<<<<<<<<<< */

/* leave the program due to error */
void
ex (char *error)
{
  fprintf (stderr, PRGNAME ": %s\n", error);
  exit (1);
}

/* called with too few/too many parameters */
void
usage (void)
{
  fprintf (stderr, PRGNAME);
#ifdef DEBUG
  fprintf (stderr, " [DEBUG mode]");
#endif
  fprintf (stderr, " version " PRGVERSION);
  fprintf (stderr, " (" PRGDATE ") ");

  data_usage ();
/*
   fprintf (stderr, "Source code available on request.\n");
 */
  fprintf (stderr,
	   "Copyleft 1999,2001,2002 by Paul David Doherty <agi@pdd.de>\n\n");
  fprintf (stderr, "Usage: " PRGNAME "      :  detect game/interpreter in current directory\n");
  fprintf (stderr, "       " PRGNAME " -ld  :  list all detected datafiles\n");
  fprintf (stderr, "       " PRGNAME " -ld2 :  list detected AGIv2 datafiles\n");
  fprintf (stderr, "       " PRGNAME " -ld3 :  list detected AGIv3 datafiles\n");
  fprintf (stderr, "       " PRGNAME " -lf  :  list detected fan-fiction datafiles\n");
  fprintf (stderr, "       " PRGNAME " -md  :  list (known) missing datafiles\n");
  exit (1);
}

void
do_filestuff (void)
{
  type16 i;

/* get length of input file */
  fseek (fdi, 0L, SEEK_END);
  infilelength = ftell (fdi);
  fseek (fdi, 0L, SEEK_SET);

  if (infilelength > 64000L)
    ex ("infile too long");

  if (infilelength == 0L)
    {
      fclose (fdi);
      return;
    }

/* load complete input file into infiledata */
  if ((infiledata = (type8 *) malloc ((type16_32) (infilelength + 2))) == NULL)
    ex ("not enough memory");
  if (fread (infiledata, (type16_32) infilelength, 1, fdi) != 1)
    ex ("couldn't finish read");

  fclose (fdi);

  crc_checksum = crc16 (infiledata, infilelength, crc_checksum);
  for (i = 0; i < infilelength; i++)
    df_checksum += (type32) infiledata[i];
}


/*>>>>>>>>>>>>>>>>>>>>> main <<<<<<<<<<<<<<<<<<<<< */

int
main (int argc, char **argv)
{
  type16 count = 0;
  type8 version;
  type16 i, j;
  type16 interp_id = 0;
  type8 unkn_int = 0, no_int = 0, unkn_gam = 0, unkn_com = 0;
  type8 match = 0, bestmatch = 0, bestmatchcount = 0;

  if (argc != 1)
    {
      if (!strcmp (argv[1], "-ld2"))
	data_list (2);
      else if (!strcmp (argv[1], "-ld3"))
	data_list (3);
      else if (!strcmp (argv[1], "-ld"))
	data_list (1);
      else if (!strcmp (argv[1], "-lf"))
	data_list (12);
      else if (!strcmp (argv[1], "-md"))
	data_list (11);
      else
	usage ();
      return (0);
    }

/* open I/O streams */
#ifdef __TURBOC__
  _fmode = O_BINARY;
#endif

/* AGI version check */

  crc_checksum = 0;

  if ((fdi = fopen ("agidata.ovl", "rb")) == NULL)
    {
/* presumably just datafiles, Mac/IIgs? */
      fprintf (stdout, "AGI interpreter: none\n");
      no_int = 1;
      fclose (fdi);
    }
  else
    {
      do_filestuff ();

      count = 0;
      version = 1;
      while (version != 0)
	{
	  version = int_table[count].crcvalue;
	  if (crc_checksum == int_table[count].crcvalue)
	    break;
	  count++;
	}

      fprintf (stdout, "AGI interpreter: ");
      if (version != 0)
	{
	  fprintf (stdout, "%s\n",
		   int_table[count].interp);
	  interp_id = int_table[count].int_id;
	}
      else
	{
	  fprintf (stdout, "unknown ");
	  unkn_int = 1;

/* find version string */
	  for (i = 0; i < (infilelength - 10); i++)
	    {
	      if ((infiledata[i] == 'V') || (infiledata[i] == 'v'))
		{
		  if ((infiledata[i + 1] == 'e') && (infiledata[i + 2] == 'r')
		      && (infiledata[i + 6] == 'n'))
		    {
		      fprintf (stdout, "(presumably: ");
		      j = i + 8;
		      while (infiledata[j] != 0)
			{
			  fprintf (stdout, "%c", infiledata[j]);
			  j++;
			}
		      fprintf (stdout, ") [CRC:%04x]", crc_checksum);
		    }
		}
	    }
	  fprintf (stdout, "\n");
	}
    }

/* game version check */

  df_checksum = 0;
  crc_checksum = 0;

  if ((fdi = fopen ("words.tok", "rb")) == NULL)
    {
/* user == stupid: give ex message, then usage! */
      fprintf (stderr, PRGNAME ": WORDS.TOK not found; no AGI game\n\n");
      usage ();
    }
  do_filestuff ();

  if ((fdi = fopen ("object", "rb")) == NULL)
    ex ("OBJECT not found; no AGI game");
  do_filestuff ();

  if ((fdi = fopen ("viewdir", "rb")) == NULL)
    {
/* presumably V3 game */
      fclose (fdi);

      if ((fdi = fopen ("grdir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
      else if ((fdi = fopen ("mhdir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
      else if ((fdi = fopen ("mh2dir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
      else if ((fdi = fopen ("kq3dir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
      else if ((fdi = fopen ("kq4dir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
      else if ((fdi = fopen ("pq1dir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
      else if ((fdi = fopen ("dmdir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
      else if ((fdi = fopen ("bcdir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
      else if ((fdi = fopen ("mgdir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
/* DIRS for Amiga MH1 */
      else if ((fdi = fopen ("dirs", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
/* fan-fic DIRs */
      else if ((fdi = fopen ("vdir", "rb")) != NULL)
	{
	  do_filestuff ();
	  fclose (fdi);
	}
      else
	ex ("DIR not found; no AGI game");
    }
  else
    {
      do_filestuff ();

      if ((fdi = fopen ("logdir", "rb")) == NULL)
	ex ("LOGDIR not found; no AGI game");
      do_filestuff ();

      if ((fdi = fopen ("picdir", "rb")) == NULL)
	ex ("PICDIR not found; no AGI game");
      do_filestuff ();

      if ((fdi = fopen ("snddir", "rb")) == NULL)
	ex ("SNDDIR not found; no AGI game");
      do_filestuff ();
    }

/* datafile version lookup with CRC & interpreter match */
  count = 0;
  version = 1;

  while (version != 0)
    {
      match = 0;
      version = version_table[count].agi_version;

      if (df_checksum == version_table[count].df_check)
	match++;
      if (interp_id == version_table[count].interp_match)
	match++;
      if (crc_checksum == version_table[count].crcvalue)
	match++;

      if ((df_checksum != version_table[count].df_check)
	  && (crc_checksum != version_table[count].crcvalue))
	match = 0;

      if (match > bestmatch)
	{
	  bestmatch = match;
	  bestmatchcount = count;
	}
      count++;
    }

/* lookup finished */
  fprintf (stdout, "AGI game: ");

  if (bestmatch != 0)
    {
      count = bestmatchcount;
      version = version_table[count].agi_version;

      if ((df_checksum != version_table[count].df_check)
	  || (crc_checksum != version_table[count].crcvalue))
	unkn_gam = 1;
      if ((interp_id != 0) && (version_table[count].interp_match != 0) &&
	  (interp_id != version_table[count].interp_match))
	unkn_com = 1;

      if (unkn_gam == 1)
	fprintf (stdout, "unknown; could be ");
      fprintf (stdout, "\"%s\" ",
	       version_table[count].game);

      if (unkn_gam == 1)
	fprintf (stdout, "[CRC:%04x/%05lx]", crc_checksum, df_checksum);
      fprintf (stdout, "\n");
    }
  else
    {
      fprintf (stdout, "unknown ");
      fprintf (stdout, "[CRC:%04x/%05lx]\n", crc_checksum, df_checksum);
      unkn_gam = 1;
    }

  fprintf (stdout, "\nResult:\n");
  if ((unkn_int == 0) && (unkn_gam == 0) && (unkn_com == 0))
    {
      fprintf (stdout, "- known game%s\n", (no_int == 0) ? " and interpreter" : "");
      if (version == FFV)
	fprintf (stdout, "- fan-fic AGIv2 game (not an official Sierra game)\n");
      if (version == FFV3)
	fprintf (stdout, "- fan-fic AGIv3 game (not an official Sierra game)\n");
    }
  else
    {
      if (unkn_int == 1)
	fprintf (stdout, "- unknown interpreter\n");
      if (unkn_gam == 1)
	fprintf (stdout, "- unknown game\n");
      if (unkn_com == 1)
	fprintf (stdout, "- unknown interpreter/game combination\n");
      fprintf (stdout, "\nPlease help me improve this program by forwarding the above\ninformation to <agi@pdd.de>\n");
    }

/* cleanup */
  free (infiledata);
  fclose (fdi);
  return (0);
}
